#include <stdio.h>
#include <sys/stat.h>
#include <unistd.h>
#include "common.h"
#include "scanner.h"
#include "task.h"
#include "util.h"

#define FMT_OK      "HTTP/1.1 %d OK\n"          \
                    "Content-Type: %s\n"        \
                    "Content-Length: %ld\n"     \
                    "Cache-Control: no-cache\n" \
                    "\n"
#define FMT_ERR     "HTTP/1.1 %d ERROR\n"

//forward declarations
static void destroy(task_t *self);
static void read_request(task_t *self);
static void write_response(task_t *self);
static void sample_response(task_t *self);
static int  get_path(task_t *req, /*out*/char *path);
static int  get_size(char *path, /*out*/long *size);
static int  get_mime_type(char *path, /*out*/char **mime_type);

//make a task
//   - fd: file descriptor for the client socket
//   - root_dir: the root directory of pages
task_t *make_task(int fd, char *root_dir) {
    static int fresh_id = 0;

    task_t *task = malloc(sizeof(task_t));
    ON_FALSE_EXIT(task != NULL, "out of memory");
    task->id = fresh_id++;
    task->fd = fd;
    task->root_dir = root_dir;
    task->method = NULL;
    task->url = NULL;

    task->read_request    = read_request;
    task->write_response  = write_response;
    task->sample_response = sample_response;
    task->destroy         = destroy;
}

//destroy self
static void destroy(task_t *self) {
    //close the socket
    close(self->fd);

    if(self->url != NULL)
        free(self->url);
    free(self);
}

//read client's request using self
static void read_request(task_t *self) {
    self->method = NULL;
    self->url = NULL;
    while(1) {
        char buf[MAX_LINE];
        int n = read_line(self->fd, buf, MAX_LINE);

        //a blank line indicates the end of the request
        if(n == 0 || buf[0] == '\n' || buf[0] == '\r')
            break;

        token_buf_t *tb = tokenize(buf);
        if(strcmp(tb->read_token(tb), "GET") == 0) {
            //TODO: set the string "GET" to method


            //TODO: set the next token value to the url
            //hint: use tb->read_token and strdup


            printf("%s", buf);
        }
        tb->destroy(tb);
    }
}

//write response to the client using self
static void write_response(task_t *self) {
    int status = 200;

    char path[MAX_LINE];
    ON_FALSE_GOTO((status = get_path(self, path)) == 200, err, "get_path failed");

    long size;
    ON_FALSE_GOTO((status = get_size(path, &size)) == 200, err, "get_size failed");

    char *mime_type;
    ON_FALSE_GOTO((status = get_mime_type(path, &mime_type)) == 200, err, "get_mime_type failed");

    //TODO: send the response header to the client
    //hint: write_str, strmsg and FMT_OK


    //TODO: send the file to the client
    //hint: use write_file


    return;
err:
    //TODO: send the status code to the client
    //hint: use write_str, strmsg and FMT_ERR

    return;
}

//send a sample response to the client
static void sample_response(task_t *self) {
    write_str(self->fd, "HTTP/1.1 200 OK\n"
                        "Content-Type: text/html\n"
                        "Content-Length: 80\n"
                        "\n"
                        "<html>"
                        "<head><title>Hello</title></head>"
                        "<body><h1>Hello World!</h1></body>"
                        "</html>");
}

//get the path to the requested file
static int get_path(task_t *req, /*out*/char *path) {
    int status = 501;   //not implemented
    ON_FALSE_GOTO(req->method != NULL, err, "unknown method");

    status = 400;       //bad request
    ON_FALSE_GOTO(req->url != NULL, err, "no URL");

    status = 200;
    char *sep = req->url[0] == '/' ? "": "/";
    sprintf(path, "%s%s%s", req->root_dir, sep, req->url);
err:
    return status;
}

//get the size of the requested file
static int get_size(char *path, /*out*/long *size) {
    int status = 404;   //not found
    struct stat st;
    //TODO: goto err if stat to path failed


    status = 403;       //forbidden
    //TODO: goto err if the path is not readable
    //hint: use access function with path and R_OK


    status = 200;
    *size = st.st_size;
err:
    return status;    
}

//get the mime type of the requested file
static int get_mime_type(char *path, /*out*/char **mime_type) {
    int status = 415;       //unsupported media type
    char *ext = strrchr(path, '.');
    ON_FALSE_GOTO(ext != NULL, err, "no extension");

    //TODO: - set mime_type for the extensions of .html, .txt, .xml,
    //        .doc, .pdf, .zip, .gif, .jpg, .jpeg, .png
    //      - for other extensions, set it to NULL


    ON_FALSE_GOTO(*mime_type != NULL, err, strmsg("unknown extension: %s", ext));
    status = 200;
err:
    return status;
}
